package org.klomp.eclipse.team.sftp;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.MessageFormat;
import java.util.Properties;

import org.apache.log4j.PropertyConfigurator;

import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IPluginDescriptor;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubProgressMonitor;

import org.eclipse.team.core.TeamException;

import org.eclipse.ui.plugin.AbstractUIPlugin;

import com.sshtools.j2ssh.configuration.ConfigurationLoader;

/**
 * The main plugin class to be used in the desktop.
 */
public class SftpPlugin extends AbstractUIPlugin
{
    public final static String ID = "org.klomp.eclipse.team.sftp";
    public final static int TRANSFER_PROGRESS_TITLE_STEPS = 100;
    public final static int TRANSFER_PROGRESS_INCREMENT = 32768;

    public final static String DEFAULT_SERVER_CMD = "/usr/libexec/openssh/sftp-server";
    public final static String LOG4J_PROPERTIES = "etc/log4j.properties";
    public final static String SSHTOOLS_CONFIG_DIR = "etc/ssh/";
    public final static boolean DEBUG = true;

    /**
     * The shared instance.
     */
    private static SftpPlugin plugin;

    /**
     * Constructor.
     */
    public SftpPlugin(IPluginDescriptor descriptor)
    {
        super(descriptor);
        plugin = this;

        initLogging();
        initJ2SshConfiguration();
    }

    private void initLogging()
    {
        Properties log4jProperties = new Properties();
        try
        {
            InputStream in = openStream(new Path(LOG4J_PROPERTIES));
            try
            {
                log4jProperties.load(in);
            }
            finally
            {
                in.close();
            }
        }
        catch (IOException e)
        {
            logError("Loading log4j configuration file", e);
        }

        // Load our own configuration
        PropertyConfigurator.configure(log4jProperties);
    }

    private void initJ2SshConfiguration()
    {
        try
        {
            URL installURL = getDescriptor().getInstallURL();
            IPath path = new Path(SSHTOOLS_CONFIG_DIR);
            URL configURL = new URL(installURL + path.toString());
            ConfigurationLoader.setHomeDir(configURL);
            ConfigurationLoader.configure();
        }
        catch (MalformedURLException e)
        {
            logError("Constructing j2ssh configuration path", e);
        }
    }

    /**
     * Returns the shared instance.
     */
    public static SftpPlugin getPlugin()
    {
        return plugin;
    }

    /**
     * Returns the workspace instance.
     */
    public static IWorkspace getWorkspace()
    {
        return ResourcesPlugin.getWorkspace();
    }

    /**
     * Returns the string from the plugin's resource bundle, or <code>key</code>
     * if not found.
     */
    public static String getString(String key)
    {
        return getPluginDescriptor().getResourceString(key);
    }

    public static String formatResourceString(String key, String arg)
    {
        MessageFormat fmt = new MessageFormat(getString(key));
        return fmt.format(new String[] { arg });
    }

    public static String formatResourceString(String key, String[] args)
    {
        MessageFormat fmt = new MessageFormat(getString(key));
        return fmt.format(args);
    }

    public static IPluginDescriptor getPluginDescriptor()
    {
        return getPlugin().getDescriptor();
    }

    public static IProgressMonitor monitorFor(IProgressMonitor monitor)
    {
        if (monitor == null)
            return new NullProgressMonitor();
        return monitor;
    }

    public static IProgressMonitor subMonitorFor(IProgressMonitor monitor, int ticks)
    {
        if (monitor == null)
            return new NullProgressMonitor();
        if (monitor instanceof NullProgressMonitor)
            return monitor;
        return new SubProgressMonitor(monitor, ticks);
    }

    /**
     * Convenience method for logging TeamExceptions to the plugin log
     */
    public static void log(TeamException e)
    {
        log(e.getStatus());
    }
    /**
     * Convenience method for logging a status to the plugin log
     */
    public static void log(IStatus status)
    {
        getPlugin().getLog().log(status);
    }

    /**
     * Convenience method for logging an internal error
     */
    public static void logError(String message, Throwable e)
    {
        log(new Status(IStatus.ERROR, ID, 0, message, e));
    }
}
