package org.klomp.eclipse.team.sftp.client;

import java.io.IOException;
import java.io.InputStream;
import java.util.Date;
import java.util.Iterator;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;

import org.eclipse.team.core.TeamException;
import org.eclipse.team.internal.core.Assert;
import org.klomp.eclipse.team.sftp.SftpException;
import org.klomp.eclipse.team.sftp.SftpPlugin;
import org.klomp.eclipse.team.sftp.target.SftpSite;
import org.klomp.eclipse.team.sftp.util.SftpProgressInputStream;
import org.klomp.sftp.packet.Attrs;
import org.klomp.sftp.packet.FxpAttrs;
import org.klomp.sftp.packet.FxpInit;
import org.klomp.sftp.packet.FxpOpen;
import org.klomp.sftp.packet.FxpRemove;
import org.klomp.sftp.packet.FxpStat;
import org.klomp.sftp.packet.FxpStatus;
import org.klomp.sftp.packet.FxpVersion;
import org.klomp.sftp.packet.SftpPacket;
import org.klomp.sftp.packet.SftpPacketFactory;
import org.klomp.sftp.protocol.Directories;
import org.klomp.sftp.protocol.PlainFile;
import org.klomp.sftp.protocol.SftpConnection;

public class SftpClient
{
    private final static int DATA_SIZE = 32000;

    private SftpServerLocation location;
    private ISftpClientListener listener;

    private SftpConnection connection;
    private SftpPacketFactory pktFactory;

    private Directories directories;

    private int timeout;

    public static final String PARENT_DIRECTORY = ".."; //$NON-NLS-1$
    public static final int USE_DEFAULT_TIMEOUT = -1;

    /**
     * Creates an SftpClient for the specified location.
     * @param location the Sftp server location
     * @param listener the Sftp listener
     * @param timeout the communications timeout in milliseconds
     */
    public SftpClient(SftpServerLocation location, ISftpClientListener listener, int timeout)
    {

        this.location = location;
        this.listener = listener;
        if (timeout == USE_DEFAULT_TIMEOUT)
        {
            this.timeout = SftpSite.DEFAULT_TIMEOUT;
        }
        else
        {
            this.timeout = timeout;
        }
    }

    /**
     * Execute the given runnable in a context that has access to an open ftp connection
     */
    public void run(String basePath, ISftpRunnable runnable, IProgressMonitor monitor) throws TeamException
    {

        if (basePath.equals(""))
            basePath = "./";
        else if (!basePath.endsWith("/"))
            basePath += "/";

        // Determine if we are nested or not
        boolean isOuterRun = false;

        try
        {
            monitor = SftpPlugin.monitorFor(monitor);
            monitor.beginTask(null, 100 + (connection == null ? 20 : 0));
            if (connection == null)
            {
                open(SftpPlugin.subMonitorFor(monitor, 10));
                isOuterRun = true;
            }
            runnable.run(basePath, SftpPlugin.subMonitorFor(monitor, 100));

        }
        finally
        {
            if (isOuterRun)
            {
                close(SftpPlugin.subMonitorFor(monitor, 10));
            }
            monitor.done();
        }
    }

    /**
     * Opens and authenticates a connection to the server, if not already open.
     * @param monitor the progress monitor, or null
     */
    public void open(IProgressMonitor monitor) throws SftpException
    {
        if (connection != null)
            return;
        monitor = SftpPlugin.monitorFor(monitor);
        monitor.beginTask(null, 100);
        monitor.subTask(SftpPlugin.formatResourceString("%SftpClient.openConnection", location.getHostname())); //$NON-NLS-1$
        try
        {
            if (location.getSshVersion() == 1)
            {
                connection =
                    new Ssh1SftpConnection(
                        location.getHostname(),
                        location.getPort(),
                        location.getServerCmd(),
                        location.getUsername(),
                        location.getPassword());
            }
            else
            {
                connection =
                    new Ssh2SftpConnection(
                        location.getHostname(),
                        location.getPort(),
                        location.getUsername(),
                        location.getPassword());
            }
            connection.setListener(listener);

            connection.open(timeout, SftpPlugin.subMonitorFor(monitor, 60));

            pktFactory = connection.getPktFactory();
            directories = new Directories(connection);
            handshake(SftpPlugin.subMonitorFor(monitor, 30));

        }
        catch (SftpException e)
        {
            if (connection != null)
            {
                try
                {
                    connection.close();
                }
                catch (IOException e2)
                {
                    throw new SftpCommunicationException("Can't close connection", e2);
                }
                monitor.worked(5);
                connection = null;
            }
            throw e;
        }
        finally
        {
            monitor.done();
        }
    }

    /**
     * Closes a connection to the server, if not already closed.
     * @param monitor the progress monitor, or null
     */
    public void close(IProgressMonitor monitor) throws SftpException
    {
        if (connection == null)
            return;
        monitor = SftpPlugin.monitorFor(monitor);
        monitor.beginTask(null, 100);
        monitor.subTask(SftpPlugin.getString("%SftpClient.closeConnection")); //$NON-NLS-1$

        try
        {
            // close the socket (50%)
            try
            {
                connection.close();
                monitor.worked(50);
            }
            catch (IOException e)
            {
                throw new SftpCommunicationException("Can't close connection", e);
            }
        }
        finally
        {
            connection = null;
            monitor.done();
        }

    }

    public boolean isConnected()
    {
        return connection != null;
    }

    /**
     * Deletes a remote directory.
     * @param directoryPath the absolute or relative path of the directory
     * @param monitor the progress monitor, or null
     */
    public void deleteDirectory(String directoryPath, IProgressMonitor monitor) throws SftpException
    {
        Assert.isNotNull(directoryPath);
        directories.deleteDirectory(directoryPath, monitor);
    }

    /**
     * Creates a remote directory.
     * @param directoryPath the absolute or relative path of the directory
     * @param monitor the progress monitor, or null
     */
    public void createDirectory(String directoryPath, IProgressMonitor monitor) throws SftpException
    {
        Assert.isNotNull(directoryPath);
        directories.createDirectory(directoryPath, monitor);
    }

    private void handshake(IProgressMonitor monitor) throws SftpException
    {
        monitor = SftpPlugin.monitorFor(monitor);
        monitor.beginTask(null, 100);
        monitor.subTask("Handshaking");
        try
        {
            FxpInit initPkt = pktFactory.createInit();
            connection.sendPacket(initPkt);
            SftpPacket response = connection.readPacket(0);
            if (response.getType() != SftpPacket.FXP_VERSION)
            {
                throw new SftpCommunicationException("Bad packet received: " + response, null);
            }
            FxpVersion versionPkt = (FxpVersion) response;
            int serverVersion = versionPkt.getVersion();
            if (serverVersion != 3 && serverVersion != 4)
            {
                throw new SftpServerException(
                    "Unsupported sftp server protocol version: " + serverVersion,
                    FxpStatus.FX_OP_UNSUPPORTED);
            }
            pktFactory.setVersion(serverVersion);
        }
        finally
        {
            monitor.done();
        }
    }

    /**
     * Deletes a remote file.
     * @param filePath the absolute or relative path of the file
     * @param monitor the progress monitor, or null
     */
    public void deleteFile(String filePath, IProgressMonitor monitor) throws SftpException
    {
        Assert.isNotNull(filePath);
        monitor = SftpPlugin.monitorFor(monitor);
        monitor.beginTask(null, 100);
        monitor.subTask(SftpPlugin.formatResourceString("%SftpClient.deleteFile", filePath)); //$NON-NLS-1$
        try
        {
            FxpRemove pkt = pktFactory.createRemove();
            pkt.setPath(filePath);
            connection.sendPacket(pkt);
            FxpStatus status = readStatus(pkt.getId());
            defaultStatusHandler(status);
            monitor.worked(50);
        }
        finally
        {
            monitor.done();
        }
    }

    private void writeFile(String filePath, InputStream localFile, IProgressMonitor monitor)
        throws SftpException
    {

        Assert.isNotNull(filePath);
        monitor = SftpPlugin.monitorFor(monitor);
        monitor.beginTask(null, 100);
        monitor.subTask(SftpPlugin.formatResourceString("%SftpClient.writeFile", filePath)); //$NON-NLS-1$

        try
        {
            PlainFile file =
                PlainFile.openFile(
                    connection,
                    filePath,
                    FxpOpen.FXF_CREAT | FxpOpen.FXF_WRITE | FxpOpen.FXF_TRUNC);
            try
            {
                file.writeFile(localFile, monitor);
            }
            finally
            {
                file.close();
            }
        }
        finally
        {
            monitor.done();
        }
    }

    private Attrs statFile(String filePath) throws SftpException
    {
        FxpStat pkt = pktFactory.createStat();
        pkt.setPath(filePath);
        pkt.setFlags(0xffffffff);
        connection.sendPacket(pkt);
        SftpPacket response = connection.readPacket(pkt.getId());

        switch (response.getType())
        {
            case SftpPacket.FXP_STATUS :
                defaultStatusHandler((FxpStatus) response);
                throw new SftpException("Unexpected non-error status received: " + response);
            case SftpPacket.FXP_ATTRS :
                FxpAttrs attrs = (FxpAttrs) response;
                return attrs.getAttrs();
            default :
                throw new SftpException("Unexpected packet received: " + response);
        }
    }

    /**
     * Retrieves the contents of a remote file.
     * 
     * The return input stream has a direct connection to the server, so
     * the contents should be read before the sftp connection is closed.
     * 
     * @param filePath the absolute or relative path of the file
     * @param binary if true, uses binary transfer type
     * @param resumeAt the position in the remote file to start at (or 0 for the whole file)
     * @param monitor the progress monitor, or null
     */
    public InputStream getContents(String filePath, boolean binary, long resumeAt, IProgressMonitor monitor)
        throws SftpException
    {
        Assert.isNotNull(filePath);

        PlainFile file = PlainFile.openFile(connection, filePath, FxpOpen.FXF_READ);

        InputStream in = file.getInputStream(resumeAt);

        //        // setup line terminator conversion
        //        if (!binary && !SftpPlugin.IS_CRLF_PLATFORM)
        //        {
        //            in = new CRLFtoLFInputStream(in);
        //        }

        return in;
    }

    /**
     * Retrieves a remote file.
     * 
     * @param filePath the absolute or relative path of the file
     * @param localFile the local file to create
     * @param binary if true, uses binary transfer type
     * @param resume if true, attempts to resume a partial transfer, else overwrites
     * @param monitor the progress monitor, or null
     */
    public void getFile(
        String filePath,
        long fileSize,
        IFile localFile,
        boolean binary,
        boolean resume,
        IProgressMonitor monitor)
        throws SftpException
    {

        monitor = SftpPlugin.monitorFor(monitor);
        monitor.beginTask(null, 100);
        final String title = SftpPlugin.formatResourceString("%SftpClient.getFile", filePath);
        monitor.subTask(title);
        try
        {
            long resumeAt = 0;
            if (resume)
            {
                resumeAt = getFileSize(localFile);
            }
            InputStream in = getContents(filePath, binary, resumeAt, SftpPlugin.subMonitorFor(monitor, 10));
            // setup progress monitoring
            IProgressMonitor subMonitor = SftpPlugin.subMonitorFor(monitor, 90);
            try
            {
                subMonitor.beginTask(null, (int) fileSize);
                subMonitor.subTask(SftpPlugin.formatResourceString("%SftpClient.transferNoSize", title));
                in = new SftpProgressInputStream(in, fileSize, title, subMonitor);

                // transfer the file
                try
                {
                    if (!localFile.exists())
                    {
                        localFile.create(in, false, null);
                    }
                    else if (resumeAt != 0)
                    {
                        // don't keep the history of the previous (incomplete) generation of the file
                        localFile.appendContents(in, false, false, null);
                    }
                    else
                    {
                        localFile.setContents(in, false, true, null);
                    }
                }
                catch (CoreException e)
                {
                    throw new SftpException("Core exception", e);
                }
            }
            finally
            {
                subMonitor.done();
            }
        }
        finally
        {
            monitor.done();
        }
    }

    /**
     * Stores a local file on the remote system.
     * @param filePath the absolute or relative path of the file
     * @param localFile the local file to send
     * @param binary if true, uses binary transfer type
     * @param monitor the progress monitor, or null
     */
    public void putFile(String filePath, IFile localFile, boolean binary, IProgressMonitor monitor)
        throws SftpException
    {

        //System.err.println("putFile: " + filePath + " from " + localFile);
        Assert.isNotNull(filePath);
        monitor = SftpPlugin.monitorFor(monitor);
        monitor.beginTask(null, 100);
        final String title = SftpPlugin.formatResourceString("%SftpClient.putFile", filePath); //$NON-NLS-1$
        monitor.subTask(title);

        try
        {
            InputStream in;
            try
            {
                in = localFile.getContents();
            }
            catch (CoreException e)
            {
                throw new SftpException("Getting contents of local file", e);
            }
            writeFile(filePath, in, monitor);
        }
        finally
        {
            monitor.done();
        }
    }

    /**
     * Lists the files stored in a remote directory.
     * @param directoryPath the absolute or relative path of the directory, or null (or empty string) for current
     * @param monitor the progress monitor, or null
     * @return an array of directory entries, or an empty array if none
     */
    public SftpDirectoryEntry[] listFiles(String directoryPath, IProgressMonitor monitor)
        throws SftpException
    {

        Assert.isNotNull(directoryPath);
        monitor = SftpPlugin.monitorFor(monitor);
        monitor.beginTask(null, 100);
        final String title = SftpPlugin.formatResourceString("%SftpClient.listFiles", directoryPath == null ? "." : directoryPath); //$NON-NLS-1$ //$NON-NLS-2$
        monitor.subTask(title);

        try
        {
            Map attrs = directories.retrieveDirectory(directoryPath, monitor);
            // don't return current and parent directory entries
            attrs.remove(".");
            attrs.remove("..");
            SftpDirectoryEntry[] entries = new SftpDirectoryEntry[attrs.size()];
            int offset = 0;
            for (Iterator i = attrs.keySet().iterator(); i.hasNext(); offset++)
            {
                String fileName = (String) i.next();

                Attrs a = (Attrs) attrs.get(fileName);

                long modMillis = 1000L * a.getModifyTime() + a.getModifyTimeNanos() / 1000L;
                boolean isDirectory = (a.getType() & Attrs.FILEXFER_TYPE_DIRECTORY) != 0;
                entries[offset] =
                    new SftpDirectoryEntry(
                        fileName,
                        isDirectory,
                        !isDirectory,
                        a.getSize(),
                        new Date(modMillis));

            }

            return entries;
        }
        finally
        {
            monitor.done();
        }
    }

    public SftpDirectoryEntry getDirectoryEntry(String filePath, IProgressMonitor monitor)
        throws SftpException
    {

        monitor = SftpPlugin.monitorFor(monitor);
        monitor.beginTask(null, 100);
        final String title = SftpPlugin.formatResourceString("%SftpClient.statFile", filePath == null ? "." : filePath); //$NON-NLS-1$ //$NON-NLS-2$
        monitor.subTask(title);

        try
        {
            Attrs attrs = statFile(filePath);

            long modMillis = 1000L * attrs.getModifyTime() + attrs.getModifyTimeNanos() / 1000L;
            
            boolean isDirectory = (attrs.getType() & Attrs.FILEXFER_TYPE_DIRECTORY) != 0;
            return new SftpDirectoryEntry(
                filePath,
                isDirectory,
                !isDirectory,
                attrs.getSize(),
                new Date(modMillis));

        }
        finally
        {
            monitor.done();
        }
    }

    private FxpStatus readStatus(int id) throws SftpException
    {
        SftpPacket packet = connection.readPacket(id);
        if (packet.getType() != SftpPacket.FXP_STATUS)
        {
            throw new SftpCommunicationException("Status packet expected but received: " + packet, null);
        }
        return (FxpStatus) packet;
    }

    /**
     * Handles a response by throwing an appropriate exception.
     * @throws SftpException if the response code is OK
     */
    private void defaultStatusHandler(FxpStatus status) throws SftpException
    {
        switch (status.getCode())
        {
            case FxpStatus.FX_OK :
                break;
            default :
                throw new SftpServerException(status.getMessage(), status.getCode());
        }
    }

    /**
     * Returns the size of an IFile.
     */
    public static long getFileSize(IFile file)
    {
        if (!file.exists())
            return 0;
        return file.getLocation().toFile().length();
    }
}
