package org.klomp.eclipse.team.sftp.target;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import org.eclipse.core.runtime.IProgressMonitor;

import org.eclipse.team.core.TeamException;
import org.klomp.eclipse.team.sftp.SftpException;
import org.klomp.eclipse.team.sftp.SftpPlugin;
import org.klomp.eclipse.team.sftp.client.ISftpRunnable;
import org.klomp.eclipse.team.sftp.util.*;

/**
 * Retrieves the contents of a remote file without writing it to the file system. Used by compare viewers.
 * There is a maximum on the size of a file that can be retrieved this way. Larger files can be downloaded
 * normally, but not with this class.
 */
public class GetContentsRunnable implements ISftpRunnable
{
    public static final int MAX_FILE_SIZE = 65535;

    private SftpRemoteTargetResource remote;

    private InputStream result;

    public GetContentsRunnable(SftpRemoteTargetResource remote)
    {
        this.remote = remote;
    }

    public void run(String basePath, IProgressMonitor monitor) throws TeamException
    {
        monitor.beginTask(null, 100);
        try
        {
            // get the size and only transfer if it is under the max file size
            int size = remote.getSize(SftpPlugin.subMonitorFor(monitor, 10));
            if (size > MAX_FILE_SIZE)
            {
                result = new ByteArrayInputStream(SftpPlugin.getString("%SftpRemoteTargetResource.remoteFileTooBig").getBytes()); //$NON-NLS-1$
                return;
            }
            final String title = SftpPlugin.formatResourceString("%SftpClient.getFile", remote.getProviderRelativePath().toString()); //$NON-NLS-1$
            monitor.subTask(title);
            InputStream in =
                remote.getProvider().getContents(
                    basePath,
                    remote.getProviderRelativePath(),
                    remote.isBinary(),
                    SftpPlugin.subMonitorFor(monitor, 10));
            // setup progress monitoring
            IProgressMonitor subMonitor = SftpPlugin.subMonitorFor(monitor, 80);
            try
            {
                subMonitor.beginTask(null, size);
                subMonitor.subTask(SftpPlugin.formatResourceString("%SftpClient.transferNoSize", title)); //$NON-NLS-1$
                in = new SftpProgressInputStream(in, size, title, subMonitor);

                // Transfer the contents
                byte[] buffer = new byte[1024];
                ByteArrayOutputStream out = new ByteArrayOutputStream();
                try
                {
                    try
                    {
                        int read;
                        while ((read = in.read(buffer)) >= 0)
                        {
                            out.write(buffer, 0, read);
                        }
                    }
                    finally
                    {
                        out.close();
                    }
                }
                finally
                {
                    in.close();
                }
                result = new ByteArrayInputStream(out.toByteArray());
            }
            catch (IOException e)
            {
                throw new SftpException("IO", e);
            }
            finally
            {
                subMonitor.done();
            }
        }
        finally
        {
            monitor.done();
        }
    }

    public InputStream getResult()
    {
        return result;
    }

}
