package org.klomp.eclipse.team.sftp.target;

import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.DateFormat;
import java.util.Locale;
import java.util.TimeZone;

import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.PlatformObject;

import org.eclipse.team.core.Team;
import org.eclipse.team.core.TeamException;
import org.eclipse.team.core.sync.IRemoteResource;
import org.eclipse.team.internal.core.target.IRemoteTargetResource;
import org.eclipse.team.internal.core.target.ITargetRunnable;
import org.eclipse.team.internal.core.target.Site;
import org.klomp.eclipse.team.sftp.SftpException;
import org.klomp.eclipse.team.sftp.SftpPlugin;
import org.klomp.eclipse.team.sftp.client.SftpDirectoryEntry;
import org.klomp.eclipse.team.sftp.client.SftpServerException;
import org.klomp.sftp.packet.FxpStatus;

public class SftpRemoteTargetResource extends PlatformObject implements IRemoteTargetResource
{
    private final static DateFormat UTC_FORMAT =
        DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.FULL, Locale.US);

    static {
        UTC_FORMAT.setTimeZone(TimeZone.getTimeZone("UTC"));
    }

    private SftpTargetProvider provider;

    private IPath providerRelativePath;

    private boolean isContainer;

    /**
     * Cache of remote information.
     */
    private SftpDirectoryEntry entry;

    public SftpRemoteTargetResource(SftpTargetProvider provider, IPath relativePath, boolean isContainer)
    {
        this(provider, relativePath, isContainer, null);
    }

    private SftpRemoteTargetResource(
        SftpTargetProvider provider,
        IPath relativePath,
        boolean container,
        SftpDirectoryEntry entry)
    {
        this.provider = provider;
        this.providerRelativePath = relativePath;
        this.isContainer = container;
        this.entry = entry;
    }

    /**
     * Delete the resource remotely. 
     * If the resource is a container, recursively delelete its members before itself.
     */
    public void delete(IProgressMonitor progress) throws TeamException
    {
        provider.run(new DeleteRecursiveRunnable(this), SftpPlugin.monitorFor(progress));
    }

    /**
     * @see IRemoteTargetResource#exists()
     */
    public boolean exists(IProgressMonitor progress) throws TeamException
    {
        entry = null;
        getEntry(progress);
        return entry != null;
    }

    /**
     * @see IRemoteResource#getContents(IProgressMonitor)
     */
    public InputStream getContents(IProgressMonitor progress) throws TeamException
    {
        GetContentsRunnable operation = new GetContentsRunnable(this);
        provider.run(operation, SftpPlugin.monitorFor(progress));
        return operation.getResult();
    }

    public boolean isBinary()
    {
        int type = Team.getType(new RemoteDisconnectedStorage(getProviderURL(), providerRelativePath));
        return type != Team.TEXT;
    }

    public String getLastModified(IProgressMonitor progress) throws TeamException
    {
        SftpDirectoryEntry entry = getEntry(progress);
        if (entry == null)
        {
            throw new SftpException(SftpPlugin.formatResourceString("%SftpRemoteTargetResource.doesNotExist", getURL().toString())); //$NON-NLS-1$
        }
        return UTC_FORMAT.format(entry.getModTime());
    }

    protected String getReleasedIdentifier(IProgressMonitor progress) throws TeamException
    {
        // First clear the entry
        entry = null;
        return getReleasedIdentifierIfNecessary(progress);
    }

    protected String getReleasedIdentifierIfNecessary(IProgressMonitor progress) throws TeamException
    {
        progress = SftpPlugin.monitorFor(progress);
        progress.beginTask(null, 100);
        try
        {
            String lastModified = getLastModified(SftpPlugin.subMonitorFor(progress, 50));
            int size = getSize(SftpPlugin.subMonitorFor(progress, 50));

            return "sftp:" + lastModified + "/" + size;
        }
        finally
        {
            progress.done();
        }
    }

    public String getName()
    {
        IPath path = providerRelativePath;
        if (path.isEmpty() || path.isRoot())
        {
            return getProviderURL().toExternalForm();
        }
        return path.lastSegment();
    }

    public SftpTargetProvider getProvider()
    {
        return provider;
    }

    public URL getProviderURL()
    {
        return provider.getURL();
    }

    public IPath getProviderRelativePath()
    {
        return providerRelativePath;
    }

    public URL getURL()
    {
        try
        {
            return new URL(getProviderURL(), providerRelativePath.toString());
        }
        catch (MalformedURLException e)
        {
            SftpPlugin.log(new TeamException(SftpPlugin.getString("%SftpRemoteTargetResource.invalidURL"), e)); //$NON-NLS-1$
            return null;
        }
    }

    public int getSize(IProgressMonitor progress) throws TeamException
    {
        SftpDirectoryEntry entry = getEntry(progress);
        if (entry == null)
        {
            throw new SftpException(SftpPlugin.formatResourceString("%SftpRemoteTargetResource.doesNotExist", getURL().toString())); //$NON-NLS-1$
        }
        return (int) entry.getSize();
    }

    public boolean isContainer()
    {
        return isContainer;
    }

    public IRemoteResource[] members(IProgressMonitor progress) throws TeamException
    {
        final SftpDirectoryEntry[] entries = provider.listFiles(getProviderRelativePath(), progress);
        IRemoteResource[] members = new IRemoteResource[entries.length];
        for (int i = 0; i < entries.length; i++)
        {
            SftpDirectoryEntry entry = entries[i];
            members[i] =
                new SftpRemoteTargetResource(
                    provider,
                    providerRelativePath.append(entry.getName()),
                    entry.isDirectory(),
                    entry);
        }
        return members;
    }

    /**
     * Create the parents of the resource remotely. If the resource is a directory, create it as well.
     */
    public void mkdirs(IProgressMonitor progress) throws TeamException
    {
        try
        {
            if (isContainer())
            {
                provider.createDirectory(getProviderRelativePath(), progress);
            }
            else
            {
                provider.createDirectory(getProviderRelativePath().removeLastSegments(1), progress);
            }
        }
        catch (SftpServerException e)
        {
            if (e.getStatus().getCode() == FxpStatus.FX_FAILURE)
            {
                return;
            }
            throw e;
        }
    }

    public IRemoteTargetResource getFile(String name)
    {
        return new SftpRemoteTargetResource(provider, providerRelativePath.append(name), false);
    }

    public IRemoteTargetResource getFolder(String name)
    {
        return new SftpRemoteTargetResource(provider, providerRelativePath.append(name), true);
    }

    public Site getSite()
    {
        return provider.getSite();
    }

    public String toString()
    {
        return getURL().toString();
    }

    private SftpDirectoryEntry getEntry(IProgressMonitor monitor) throws TeamException
    {
        if (entry == null)
        {
            if (isContainer())
            {
                entry = provider.fetchEntry(getProviderRelativePath(), monitor);
            }
            else
            {
                entry = provider.fetchEntryForFile(getProviderRelativePath(), monitor);
            }
        }
        return entry;
    }

    public boolean canBeReached(IProgressMonitor monitor) throws TeamException
    {
        // Doing something, even "nothing", will trigger a connection.
        provider.run(new ITargetRunnable()
        {
            public void run(IProgressMonitor monitor) throws TeamException
            {
            }
        }, monitor);
        // We connected with no exception
        return true;
    }

    public boolean equals(Object o)
    {
        if (o == null || o.getClass() != getClass())
        {
            return false;
        }

        if (o == this)
        {
            return true;
        }

        SftpRemoteTargetResource other = (SftpRemoteTargetResource) o;
        return other.getURL().equals(getURL());
    }

    public int hashCode()
    {
        return getURL().hashCode();
    }

}
