package org.klomp.eclipse.team.sftp.target;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Properties;

import org.eclipse.core.runtime.IPath;

import org.eclipse.team.core.TeamException;
import org.eclipse.team.internal.core.Assert;
import org.eclipse.team.internal.core.target.AuthenticatedSite;
import org.eclipse.team.internal.core.target.TargetProvider;
import org.klomp.eclipse.team.sftp.SftpPlugin;

public class SftpSite extends AuthenticatedSite
{
    public final static int DEFAULT_TIMEOUT = 60;
    public final static int DEFAULT_PORT = 22;

    private String host;
    private int port;
    private String serverCmd;
    private String basePath;
    private int timeout = DEFAULT_TIMEOUT;
    private int sshVersion = 2;

    public SftpSite(Properties configuration)
    {
        try
        {
            configure(configuration);
        }
        catch (TeamException e)
        {
            SftpPlugin.log(e.getStatus());
        }
    }

    public SftpSite(ObjectInputStream is)
    {
        Properties properties = new Properties();

        try
        {
            is.readInt(); // versionId
            String propertiesString = is.readUTF();
            ByteArrayInputStream inputStream = new ByteArrayInputStream(propertiesString.getBytes());

            // Read the properties from the stream.
            properties.load(inputStream);
        }
        catch (IOException e)
        {
            SftpPlugin.logError(SftpPlugin.getString("%Config.error"), e); //$NON-NLS-1$
        }

        try
        {
            configure(properties);
        }
        catch (TeamException e)
        {
            SftpPlugin.logError(SftpPlugin.getString("%Config.error"), e); //$NON-NLS-1$
        }
    }

    public Properties getConfiguration()
    {
        Properties properties = new Properties();
        properties.setProperty("host", this.host); //$NON-NLS-1$
        properties.setProperty("port", String.valueOf(this.port)); //$NON-NLS-1$
        properties.setProperty("sshVersion", Integer.toString(this.sshVersion)); // $NON-NLS-1$
        properties.setProperty("basePath", this.basePath); //$NON-NLS-1$
        properties.setProperty("serverCmd", this.serverCmd); //$NON-NLS-1$
        properties.setProperty("username", this.username); //$NON-NLS-1$
        properties.setProperty("password", this.password); //$NON-NLS-1$
        properties.setProperty("connectionTimeout", Integer.toString(this.timeout)); //$NON-NLS-1$
        return properties;
    }

    private void configure(Properties configuration) throws TeamException
    {
        Assert.isNotNull(configuration, "configuration"); //$NON-NLS-1$

        // Location.
        this.host = configuration.getProperty("host"); //$NON-NLS-1$

        String portString = configuration.getProperty("port"); //$NON-NLS-1$
        if (portString != null && portString.length() > 0)
        {
            try
            {
                this.port = Integer.parseInt(portString);
            }
            catch (NumberFormatException ex)
            {
                throw new TeamException(SftpPlugin.getString("%SftpSite.Exception_converting_port_value_2"), ex); //$NON-NLS-1$
            }
        }

        String sshVersionString = configuration.getProperty("sshVersion"); // $NON_NLS-1$
        if (sshVersionString == null || "1".equals(sshVersionString))
        {
            // Older versions of the plugin didn't write this configuration property,
            // and these versions only supported ssh version 1.
            this.sshVersion = 1;
        }
        else
        {
            this.sshVersion = 2;
        }

        this.serverCmd = configuration.getProperty("serverCmd"); //$NON-NLS-1$
        if (this.serverCmd == null)
            this.serverCmd = SftpPlugin.DEFAULT_SERVER_CMD;

        this.basePath = configuration.getProperty("basePath"); //$NON-NLS-1$

        // Authentication.
        this.username = configuration.getProperty("username"); //$NON-NLS-1$
        this.password = configuration.getProperty("password"); //$NON-NLS-1$

        // Timeouts.
        String timeoutString = configuration.getProperty("connectionTimeout"); //$NON-NLS-1$
        if (timeoutString != null && timeoutString.length() > 0)
        {
            try
            {
                timeout = Integer.parseInt(timeoutString);
            }
            catch (NumberFormatException ex)
            {
                throw new TeamException(SftpPlugin.getString("%SftpSite.Exception_converting_timeout_value_2"), ex); //$NON-NLS-1$
            }
        }
        initURLForm();
    }

    private void initURLForm() throws TeamException
    {
        String path;
        if (basePath.startsWith("/"))
        {
            path = basePath + "/";
        }
        else if (basePath.length() == 0)
        {
            path = "/~/";
        }
        else
        {
            path = "/~/" + basePath + "/";
        }

        try
        {
            if (port == SftpSite.DEFAULT_PORT)
            {
                setURL(new URL("sftp" + sshVersion, username + "@" + host, path));
            }
            else
            {
                setURL(new URL("sftp" + sshVersion, username + "@" + host, port, path));
            }
        }
        catch (MalformedURLException e)
        {
            throw new TeamException(SftpPlugin.getString("%SftpSite.Exception_creating_url"), e); //$NON-NLS-1$
        }
    }

    /**
     * @see Site#newProvider(IPath)
     */
    public TargetProvider newProvider(IPath intrasitePath) throws TeamException
    {
        return new SftpTargetProvider(this, intrasitePath);
    }

    /**
     * @see Site#getType()
     */
    public String getType()
    {
        return SftpSiteFactory.TARGET_TYPE;
    }

    /**
     * Gets the timeout in seconds.
     * @return Returns a int
     */
    int getTimeout()
    {
        return timeout;
    }

    /**
     * @see Site#writeObject(ObjectOutputStream)
     */
    public void writeObject(ObjectOutputStream os)
    {

        Properties configuration = getConfiguration();
        // Flatten the configuration to bytes.
        ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
        try
        {
            configuration.store(byteStream, null);
            byteStream.close();

            os.writeInt(SftpSiteFactory.CONFIG_FORMAT_VERSION);
            os.writeUTF(byteStream.toString());
            os.flush();
        }
        catch (IOException e)
        {
            SftpPlugin.logError(SftpPlugin.getString("%SftpSite.Exception_writing_configuration_to_stream_3"), e); //$NON-NLS-1$
        }
    }

    public void dispose() throws TeamException
    {
    }

    public String getServerCmd()
    {
        return serverCmd;
    }

    public int getSshVersion()
    {
        return sshVersion;
    }

    public void setSshVersion(int sshVersion)
    {
        this.sshVersion = sshVersion;
    }

}
