/*******************************************************************************
 * Copyright (c) 2002 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Common Public License v0.5
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v05.html
 * 
 * Contributors:
 * IBM - Initial API and implementation
 * Erwin Bolwidt - SFTP modifications
 ******************************************************************************/
package org.klomp.eclipse.team.sftp.target;

import java.io.InputStream;
import java.net.URL;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;

import org.eclipse.team.core.Team;
import org.eclipse.team.core.TeamException;
import org.eclipse.team.core.sync.IRemoteSyncElement;
import org.eclipse.team.internal.core.target.IRemoteTargetResource;
import org.eclipse.team.internal.core.target.ITargetRunnable;
import org.eclipse.team.internal.core.target.ResourceState;
import org.eclipse.team.internal.core.target.SynchronizedTargetProvider;
import org.eclipse.team.internal.core.target.UrlUtil;
import org.klomp.eclipse.team.sftp.SftpException;
import org.klomp.eclipse.team.sftp.SftpPlugin;
import org.klomp.eclipse.team.sftp.client.ISftpClientListener;
import org.klomp.eclipse.team.sftp.client.ISftpRunnable;
import org.klomp.eclipse.team.sftp.client.SftpClient;
import org.klomp.eclipse.team.sftp.client.SftpDirectoryEntry;
import org.klomp.eclipse.team.sftp.client.SftpServerLocation;
import org.klomp.eclipse.team.sftp.util.*;
import org.klomp.sftp.packet.FxpStatus;

public final class SftpTargetProvider extends SynchronizedTargetProvider
{
    private static Map openClients = Collections.synchronizedMap(new HashMap());

    private static final SftpDirectoryEntry IS_URL_ROOT = new SftpDirectoryEntry(null, true, false, 0, null);

    public SftpTargetProvider(SftpSite site, IPath intrasitePath) throws TeamException
    {
        super(site, intrasitePath.makeRelative());
    }

    /**
     * @see TargetProvider#run(ITargetRunnable, IProgressMonitor)
     */
    public void run(final ITargetRunnable runnable, IProgressMonitor monitor) throws TeamException
    {
        run(new ISftpRunnable()
        {
            public void run(String basePath, IProgressMonitor monitor) throws TeamException
            {
                SftpTargetProvider.super.run(runnable, monitor);
            }
        }, monitor);
    }

    /**
     * Execute the given runnable in a context that has access to an open ftp connection
     * that is opened in the directory of the provider.
     */
    public void run(ISftpRunnable runnable, IProgressMonitor monitor) throws TeamException
    {
        SftpClient client = getClient();

        monitor = SftpPlugin.monitorFor(monitor);
        try
        {
            monitor.beginTask(null, 110 + (client.isConnected() ? 0 : 20));
            String basePath = getBasePath();

            if (client.isConnected())
            {
                client.run(basePath, runnable, SftpPlugin.subMonitorFor(monitor, 100));
            }
            else
            {
                client.open(SftpPlugin.subMonitorFor(monitor, 10));
                openClients.put(getURL(), client);
                try
                {
                    client.run(basePath, runnable, SftpPlugin.subMonitorFor(monitor, 100));
                }
                finally
                {
                    openClients.remove(getURL());
                    client.close(SftpPlugin.subMonitorFor(monitor, 10));
                }
            }
        }
        finally
        {
            monitor.done();
        }
    }

    protected String getBasePath()
    {
        URL url = getURL();
        String path = url.getPath();
        if (path.startsWith("/~/"))
        {
            return path.substring(3);
        }
        else
        {
            return path;
        }
    }

    public SftpClient getClient()
    {
        SftpClient client = (SftpClient) openClients.get(getURL());
        final SftpSite site = (SftpSite) getSite();
        if (client != null)
        {
            return client;
        }

        // Authentication.
        String username = site.getUsername();
        String password = site.getPassword();
        SftpServerLocation location =
            new SftpServerLocation(getURL(), site.getSshVersion(), site.getServerCmd(), username, password);

        ISftpClientListener listener;

        if (SftpPlugin.DEBUG)
        {
            listener = new DebugSftpClientListener(site);
        }
        else
        {
            listener = null;
        }
        // Create client
        return new SftpClient(location, listener, site.getTimeout() * 1000);
    }

    /**
    * @see SynchronizedTargetProvider#newState(IResource)
    */
    public ResourceState newState(IResource resource)
    {
        return new SftpResourceState(this, resource);
    }

    /**
     * @see SynchronizedTargetProvider#newState(IResource, IRemoteTargetResource)
     */
    public ResourceState newState(IResource resource, IRemoteTargetResource remote)
    {
        // The provider of the given remote resource must match the receiver.
        if (remote != null)
        {
            SftpRemoteTargetResource sftpRemote = (SftpRemoteTargetResource) remote;
            SftpTargetProvider provider = sftpRemote.getProvider();
            if (!intrasitePath.equals(provider.intrasitePath))
            {
                sftpRemote =
                    new SftpRemoteTargetResource(
                        this,
                        UrlUtil.getTrailingPath(remote.getURL(), targetURL),
                        resource.getType() != IResource.FILE);
            }
            return new SftpResourceState(resource, sftpRemote);
        }
        else
        {
            return newState(resource);
        }
    }

    /**
     * @see TargetProvider#getRemoteResource()
     */
    public IRemoteTargetResource getRemoteResource()
    {
        final boolean IS_CONTAINER = true;
        return new SftpRemoteTargetResource(this, Path.EMPTY, IS_CONTAINER);
    }

    /**
     * @see TargetProvider#getRemoteResourceFor(IResource)
     */
    public IRemoteTargetResource getRemoteResourceFor(IResource resource)
    {
        return new SftpRemoteTargetResource(
            this,
            resource.getProjectRelativePath(),
            resource.getType() != IResource.FILE);
    }

    /**
     * @see TargetProvider#getRemoteSyncElement(IResource)
     */
    public IRemoteSyncElement getRemoteSyncElement(IResource resource)
    {
        return new SftpRemoteSyncElement(this, resource, getRemoteResourceFor(resource));
    }

    protected void createDirectory(final IPath providerRelativePath, IProgressMonitor monitor)
        throws TeamException
    {

        if (providerRelativePath.isEmpty())
        {
            return;
        }
        run(new CreateDirectoriesRunnable(this, providerRelativePath), SftpPlugin.monitorFor(monitor));

    }

    protected void deleteFile(String basePath, IPath providerRelativePath, IProgressMonitor monitor)
        throws SftpException
    {
        getClient().deleteFile(basePath + providerRelativePath.toString(), monitor);
    }

    protected void deleteDirectory(String basePath, IPath providerRelativePath, IProgressMonitor monitor)
        throws SftpException
    {
        getClient().deleteDirectory(basePath + providerRelativePath.toString(), monitor);
    }

    /*
     * This method probably throws an exception if the parent doesn't exist
     */
    protected SftpDirectoryEntry fetchEntry(IPath providerRelativePath, IProgressMonitor progress)
        throws TeamException
    {

        try
        {
            SftpDirectoryEntry[] entries = listFiles(providerRelativePath.removeLastSegments(1), progress);
            if (providerRelativePath.isEmpty())
            {
                // We're at the root and it must exist since we got entries
                return IS_URL_ROOT;
            }
            else
            {
                for (int i = 0; i < entries.length; i++)
                {
                    SftpDirectoryEntry dirEntry = entries[i];
                    if (dirEntry.getName().equals(providerRelativePath.lastSegment()))
                    {
                        return dirEntry;
                    }
                }
            }
            return null;
        }
        catch (TeamException e)
        {
            final int code = e.getStatus().getCode();
            if (code == FxpStatus.FX_NO_SUCH_FILE || code == FxpStatus.FX_NO_SUCH_PATH)
            {
                return null;
            }
            else
            {
                throw e;
            }
        }
    }

    /*
     * This method probably throws an exception if the parent doesn't exist
     */
    protected SftpDirectoryEntry fetchEntryForFile(IPath providerRelativePath, IProgressMonitor progress)
        throws TeamException
    {
        try
        {
            return getDirectoryEntry(providerRelativePath, progress);
        }
        catch (TeamException e)
        {
            final int code = e.getStatus().getCode();
            if (code == FxpStatus.FX_NO_SUCH_FILE || code == FxpStatus.FX_NO_SUCH_PATH)
            {
                return null;
            }
            else
            {
                throw e;
            }
        }
    }

    protected InputStream getContents(
        String basePath,
        IPath providerRelativePath,
        boolean isBinary,
        IProgressMonitor monitor)
        throws SftpException
    {
        return getClient().getContents(basePath + providerRelativePath.toString(), isBinary, 0L, monitor);
    }

    protected void getFile(String basePath, long size, IFile localFile, IProgressMonitor monitor) throws SftpException
    {
        final String remotePath = basePath + localFile.getProjectRelativePath().toString();
        getClient().getFile(remotePath, size, localFile, Team.getType(localFile) != Team.TEXT, false, monitor);
    }

    protected void putFile(String basePath, IFile localFile, IProgressMonitor monitor) throws SftpException
    {
        final String remotePath = basePath + localFile.getProjectRelativePath().toString();
        getClient().putFile(remotePath, localFile, Team.getType(localFile) != Team.TEXT, monitor);
    }

    protected SftpDirectoryEntry[] listFiles(final IPath providerRelativePath, IProgressMonitor monitor)
        throws TeamException
    {
        final SftpDirectoryEntry[][] entries = new SftpDirectoryEntry[1][0];
        run(new ISftpRunnable()
        {
            public void run(String basePath, IProgressMonitor monitor) throws TeamException
            {
                String directory = basePath;
                if (!providerRelativePath.isEmpty())
                {
                    directory += providerRelativePath.toString();
                }
                entries[0] = getClient().listFiles(directory, monitor);
            }
        }, SftpPlugin.monitorFor(monitor));
        return entries[0];
    }

    protected SftpDirectoryEntry getDirectoryEntry(
        final IPath providerRelativePath,
        IProgressMonitor monitor)
        throws TeamException
    {
        final SftpDirectoryEntry[] entry = new SftpDirectoryEntry[1];
        run(new ISftpRunnable()
        {
            public void run(String basePath, IProgressMonitor monitor) throws TeamException
            {
                String directory = basePath;
                if (!providerRelativePath.isEmpty())
                {
                    directory += providerRelativePath.toString();
                }
                entry[0] = getClient().getDirectoryEntry(directory, monitor);
            }
        }, SftpPlugin.monitorFor(monitor));
        return entry[0];
    }
}
