package org.klomp.eclipse.team.sftp.ui;

import java.lang.reflect.InvocationTargetException;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;

import org.eclipse.team.internal.core.target.Site;
import org.eclipse.team.internal.core.target.TargetManager;
import org.eclipse.team.internal.ui.UIConstants;
import org.eclipse.team.internal.ui.target.ConfigureTargetWizard;
import org.eclipse.team.internal.ui.target.MappingSelectionPage;
import org.eclipse.team.ui.IConfigurationWizard;
import org.eclipse.team.ui.TeamImages;

import org.eclipse.ui.IWorkbench;

import org.klomp.eclipse.team.sftp.SftpPlugin;
import org.klomp.eclipse.team.sftp.target.SftpSite;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.IDialogSettings;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.wizard.IWizardPage;
import org.eclipse.jface.wizard.Wizard;

public class SftpConfigurationWizard extends Wizard implements IConfigurationWizard
{

    private IProject project;
    private SftpMainPage mainPage;

    public SftpConfigurationWizard()
    {
        // initializes the dialogs settings store
        IDialogSettings workbenchSettings = SftpPlugin.getPlugin().getDialogSettings();
        IDialogSettings section = workbenchSettings.getSection("SftpConfigurationWizard"); //$NON-NLS-1$
        if (section == null)
        {
            section = workbenchSettings.addNewSection("SftpConfigurationWizard"); //$NON-NLS-1$
        }

        setDialogSettings(section);
    }

    protected String getMainPageTitle()
    {
        return SftpPlugin.getString("%SftpMainPage.name"); //$NON-NLS-1$
    }

    protected String getMainPageDescription()
    {
        return SftpPlugin.getString("%SftpMainPage.description"); //$NON-NLS-1$
    }

    public void addPages()
    {
        mainPage = new SftpMainPage("sftp-site-creation", getMainPageTitle(), getMainPageDescription(), TeamImages.getImageDescriptor(UIConstants.IMG_WIZBAN_SHARE)); //$NON-NLS-1$
        mainPage.setDialogSettings(getDialogSettings());
        addPage(mainPage);
    }

    /**
     * Finished pressed while on the target pages. If finished is pressed while on the mapping
     * page, the target configuration wizard will handle this case.
     */
    public boolean performFinish()
    {
        mainPage.finish(null);
        SftpSite site = mainPage.getSite();
        if (ConfigureTargetWizard.validateSite(site, getContainer()))
        {
            TargetManager.addSite(site);
            return true;
        }
        else
        {
            return false;
        }
    }

    public void init(IWorkbench workbench, IProject project)
    {
        this.project = project;
    }

    /*
     * Convenience method for running an operation with progress and
     * error feedback.
     * 
     * @param runnable  the runnable to run
     * @param problemMessage an optional message to display in case of errors
     */
    protected void run(IRunnableWithProgress runnable, String problemMessage)
    {
        IStatus errors = null;
        boolean internalError = false;
        try
        {
            new ProgressMonitorDialog(getShell()).run(false, false, runnable);
        }
        catch (InvocationTargetException e)
        {
            Throwable t = e.getTargetException();
            if (t instanceof CoreException)
            {
                errors = ((CoreException) t).getStatus();
            }
            else
            {
                errors = new Status(IStatus.ERROR, SftpPlugin.ID, 1, problemMessage, t);
                internalError = true;
            }
        }
        catch (InterruptedException e)
        {
            errors = null;
        }
        if (errors != null)
        {
            String msg =
                internalError
                    ? SftpPlugin.getString("%SftpMainPage.internalError")
                    : problemMessage;
            ErrorDialog.openError(getShell(), msg, null, errors);
        }
    }

    public boolean canFinish()
    {
        IWizardPage page = getContainer().getCurrentPage();
        if (page == mainPage && project != null)
        {
            return false;
        }
        return super.canFinish();
    }

    public IWizardPage getNextPage(IWizardPage page)
    {
        if (page == mainPage)
        {
            Site site = mainPage.getSite();
            if (site != null && project != null)
            {
                MappingSelectionPage mappingPage = ConfigureTargetWizard.getMappingPage();
                mappingPage.setSite(site);
                mappingPage.setPreviousPage(page);
                return mappingPage;
            }
        }
        return super.getNextPage(page);
    }

    public String getWindowTitle()
    {
        return SftpPlugin.getString("%SftpMainPage.windowTitle"); //$NON-NLS-1$
    }
}