package org.klomp.sftp.packet;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import org.klomp.ssh.DataIO;

public class FxpStatus extends SftpPacket
{
    /**
     * Indicates successful completion of the operation.
     */
    public final static int FX_OK = 0;
    /**
     * Indicates end-of-file condition; for SSH_FX_READ it means that no
     * more data is available in the file, and for SSH_FX_READDIR it
     * indicates that no more files are contained in the directory.
     */
    public final static int FX_EOF = 1;
    /**
     * Is returned when a reference is made to a file which does not
     * exist.
     */
    public final static int FX_NO_SUCH_FILE = 2;
    /**
     * Is returned when the authenticated user does not have sufficient
     * permissions to perform the operation.
     */
    public final static int FX_PERMISSION_DENIED = 3;
    /**
     * Is a generic catch-all error message; it should be returned if an
     * error occurs for which there is no more specific error code
     * defined.
     */
    public final static int FX_FAILURE = 4;
    /**
     * May be returned if a badly formatted packet or protocol
     * incompatibility is detected.
     */
    public final static int FX_BAD_MESSAGE = 5;
    /**
     * Is a pseudo-error which indicates that the client has no
     * connection to the server (it can only be generated locally by the
     * client, and MUST NOT be returned by servers).
     */
    public final static int FX_NO_CONNECTION = 6;
    /**
     * Is a pseudo-error which indicates that the connection to the
     * server has been lost (it can only be generated locally by the
     * client, and MUST NOT be returned by servers).
     */
    public final static int FX_CONNECTION_LOST = 7;
    /**
     * Indicates that an attempt was made to perform an operation which
     * is not supported for the server (it may be generated locally by
     * the client if e.g.  the version number exchange indicates that a
     * required feature is not supported by the server, or it may be
     * returned by the server if the server does not implement an
     * operation).
     */
    public final static int FX_OP_UNSUPPORTED = 8;
    /**
     * The handle value was invalid.
     */
    public final static int FX_INVALID_HANDLE = 9;
    /**
     * The file path does not exist or is invalid.
     */
    public final static int FX_NO_SUCH_PATH = 10;
    /**
     * The file already exists.
     */
    public final static int FX_FILE_ALREADY_EXISTS = 11;
    /**
     * The file is on read only media, or the media is write protected.
     */
    public final static int FX_WRITE_PROTECT = 12;
    /**
     * The requested operation can not be completed because there is no
     * media available in the drive.
     */
    public final static int FX_NO_MEDIA = 13;

    /**
     * Status code received in this message.
     */
    private int code;
    /**
     * Human-readable explanation of status.
     */
    private String message;
    /**
     * Language in which above message is written.
     */
    private String language;

    FxpStatus(int version)
    {
        super(SftpPacket.FXP_STATUS, version);
    }

    FxpStatus(int version, int id)
    {
        super(SftpPacket.FXP_STATUS, version, id);
    }

    protected void readPayload(InputStream payload) throws IOException
    {
        id = DataIO.readInt(payload);
        code = DataIO.readInt(payload);
        message = DataIO.readStringUTF8(payload);
        language = DataIO.readStringAscii(payload);

    }

    protected void writePayload(ByteArrayOutputStream payload)
    {
        DataIO.writeInt(payload, id);
        DataIO.writeInt(payload, code);
        DataIO.writeStringUTF8(payload, message);
        DataIO.writeStringAscii(payload, language);
    }

    public boolean isNoSuchFileOrPath()
    {
        return code == FX_NO_SUCH_FILE || code == FX_NO_SUCH_PATH;
    }

    public boolean isOK()
    {
        return code == FX_OK;
    }

    public boolean isFailure()
    {
        return code == FX_FAILURE;
    }

    public int getCode()
    {
        return code;
    }

    public String getLanguage()
    {
        return language;
    }

    public String getMessage()
    {
        return message;
    }

    public void setCode(int code)
    {
        this.code = code;
    }

    public void setLanguage(String language)
    {
        this.language = language;
    }

    public void setMessage(String message)
    {
        this.message = message;
    }

}