package org.klomp.sftp.packet;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;

import org.klomp.ssh.DataIO;

public class SftpPacketFactory
{
    public final static int VERSION_3 = 3;
    public final static int VERSION_4 = 4;

    /**
     * Protocol version used when creating packets. Packets write/read themselves differently depending
     * on the protocol version.
     */
    private int version;

    private int nextId = 1;

    private SftpPacketFactory(int version)
    {
        this.version = version;
    }

    public static SftpPacketFactory create(int version)
    {
        return new SftpPacketFactory(version);
    }

    public int getVersion()
    {
        return version;
    }

    public void setVersion(int version)
    {
        this.version = version;
    }

    public SftpPacket readPacket(InputStream in) throws IOException
    {
        int payloadLength = DataIO.readInt(in);

        payloadLength -= 1;
        // type field length is in the packet but not in the java classes.
        byte[] payload = new byte[payloadLength];
        int type = in.read();
        DataIO.readFully(in, payload);

        SftpPacket packet;
        switch (type)
        {
            case SftpPacket.FXP_INIT :
                packet = new FxpInit(version);
                break;
            case SftpPacket.FXP_VERSION :
                packet = new FxpVersion(version);
                break;
            case SftpPacket.FXP_OPEN :
                packet = new FxpOpen(version);
                break;
            case SftpPacket.FXP_CLOSE :
                packet = new FxpClose(version);
                break;
            case SftpPacket.FXP_READ :
                packet = new FxpRead(version);
                break;
            case SftpPacket.FXP_WRITE :
                packet = new FxpWrite(version);
                break;
            case SftpPacket.FXP_LSTAT :
                packet = new FxpLstat(version);
                break;
            case SftpPacket.FXP_FSTAT :
                packet = new FxpFstat(version);
                break;
            case SftpPacket.FXP_SETSTAT :
                packet = new FxpSetStat(version);
                break;
            case SftpPacket.FXP_FSETSTAT :
                packet = new FxpFsetStat(version);
                break;
            case SftpPacket.FXP_OPENDIR :
                packet = new FxpOpenDir(version);
                break;
            case SftpPacket.FXP_READDIR :
                packet = new FxpReadDir(version);
                break;
            case SftpPacket.FXP_REMOVE :
                packet = new FxpRemove(version);
                break;
            case SftpPacket.FXP_MKDIR :
                packet = new FxpMkDir(version);
                break;
            case SftpPacket.FXP_RMDIR :
                packet = new FxpRmDir(version);
                break;
            case SftpPacket.FXP_REALPATH :
                packet = new FxpRealPath(version);
                break;
            case SftpPacket.FXP_STAT :
                packet = new FxpStat(version);
                break;
            case SftpPacket.FXP_RENAME :
                packet = new FxpRename(version);
                break;
            case SftpPacket.FXP_READLINK :
                packet = new FxpReadLink(version);
                break;
            case SftpPacket.FXP_SYMLINK :
                packet = new FxpSymLink(version);
                break;
            case SftpPacket.FXP_STATUS :
                packet = new FxpStatus(version);
                break;
            case SftpPacket.FXP_HANDLE :
                packet = new FxpHandle(version);
                break;
            case SftpPacket.FXP_DATA :
                packet = new FxpData(version);
                break;
            case SftpPacket.FXP_NAME :
                packet = new FxpName(version);
                break;
            case SftpPacket.FXP_ATTRS :
                packet = new FxpAttrs(version);
                break;
            case SftpPacket.FXP_EXTENDED :
                packet = new FxpExtended(version);
                break;
            case SftpPacket.FXP_EXTENDED_REPLY :
                packet = new FxpExtendedReply(version);
                break;
            default :
                throw new IOException("Unknown packet type: " + type);
        }
        ByteArrayInputStream payloadIn = new ByteArrayInputStream(payload);
        packet.readPayload(payloadIn);
        return packet;
    }

    public SftpPacket createPacket(int type)
    {
        SftpPacket packet;
        switch (type)
        {
            case SftpPacket.FXP_INIT :
                packet = new FxpInit(version);
                break;
            case SftpPacket.FXP_VERSION :
                packet = new FxpVersion(version);
                break;
            case SftpPacket.FXP_OPEN :
                packet = new FxpOpen(version, nextId++);
                break;
            case SftpPacket.FXP_CLOSE :
                packet = new FxpClose(version, nextId++);
                break;
            case SftpPacket.FXP_READ :
                packet = new FxpRead(version, nextId++);
                break;
            case SftpPacket.FXP_WRITE :
                packet = new FxpWrite(version, nextId++);
                break;
            case SftpPacket.FXP_LSTAT :
                packet = new FxpLstat(version, nextId++);
                break;
            case SftpPacket.FXP_FSTAT :
                packet = new FxpFstat(version, nextId++);
                break;
            case SftpPacket.FXP_SETSTAT :
                packet = new FxpSetStat(version, nextId++);
                break;
            case SftpPacket.FXP_FSETSTAT :
                packet = new FxpFsetStat(version, nextId++);
                break;
            case SftpPacket.FXP_OPENDIR :
                packet = new FxpOpenDir(version, nextId++);
                break;
            case SftpPacket.FXP_READDIR :
                packet = new FxpReadDir(version, nextId++);
                break;
            case SftpPacket.FXP_REMOVE :
                packet = new FxpRemove(version, nextId++);
                break;
            case SftpPacket.FXP_MKDIR :
                packet = new FxpMkDir(version, nextId++);
                break;
            case SftpPacket.FXP_RMDIR :
                packet = new FxpRmDir(version, nextId++);
                break;
            case SftpPacket.FXP_REALPATH :
                packet = new FxpRealPath(version, nextId++);
                break;
            case SftpPacket.FXP_STAT :
                packet = new FxpStat(version, nextId++);
                break;
            case SftpPacket.FXP_RENAME :
                packet = new FxpRename(version, nextId++);
                break;
            case SftpPacket.FXP_READLINK :
                packet = new FxpReadLink(version, nextId++);
                break;
            case SftpPacket.FXP_SYMLINK :
                packet = new FxpSymLink(version, nextId++);
                break;
            case SftpPacket.FXP_STATUS :
                packet = new FxpStatus(version, nextId++);
                break;
            case SftpPacket.FXP_HANDLE :
                packet = new FxpHandle(version, nextId++);
                break;
            case SftpPacket.FXP_DATA :
                packet = new FxpData(version, nextId++);
                break;
            case SftpPacket.FXP_NAME :
                packet = new FxpName(version, nextId++);
                break;
            case SftpPacket.FXP_ATTRS :
                packet = new FxpAttrs(version, nextId++);
                break;
            case SftpPacket.FXP_EXTENDED :
                packet = new FxpExtended(version, nextId++);
                break;
            case SftpPacket.FXP_EXTENDED_REPLY :
                packet = new FxpExtendedReply(version, nextId++);
                break;
            default :
                throw new IllegalArgumentException("Bad packet type: " + type);
        }
        return packet;
    }

    public FxpInit createInit()
    {
        return new FxpInit(version);
    }
    public FxpVersion createVersion()
    {
        return new FxpVersion(version);
    }

    public FxpOpen createOpen()
    {
        return new FxpOpen(version, nextId++);
    }

    public FxpClose createClose()
    {
        return new FxpClose(version, nextId++);
    }

    public FxpRead createRead()
    {
        return new FxpRead(version, nextId++);
    }

    public FxpWrite createWrite()
    {
        return new FxpWrite(version, nextId++);
    }

    public FxpLstat createLstat()
    {
        return new FxpLstat(version, nextId++);
    }

    public FxpFstat createFstat()
    {
        return new FxpFstat(version, nextId++);
    }

    public FxpSetStat createSetStat()
    {
        return new FxpSetStat(version, nextId++);
    }

    public FxpFsetStat createFsetStat()
    {
        return new FxpFsetStat(version, nextId++);
    }

    public FxpOpenDir createOpenDir()
    {
        return new FxpOpenDir(version, nextId++);
    }

    public FxpReadDir createReadDir()
    {
        return new FxpReadDir(version, nextId++);
    }

    public FxpRemove createRemove()
    {
        return new FxpRemove(version, nextId++);
    }

    public FxpMkDir createMkDir()
    {
        return new FxpMkDir(version, nextId++);
    }

    public FxpRmDir createRmDir()
    {
        return new FxpRmDir(version, nextId++);
    }

    public FxpRealPath createRealPath()
    {
        return new FxpRealPath(version, nextId++);
    }

    public FxpStat createStat()
    {
        return new FxpStat(version, nextId++);
    }

    public FxpRename createRename()
    {
        return new FxpRename(version, nextId++);
    }

    public FxpReadLink createReadLink()
    {
        return new FxpReadLink(version, nextId++);
    }

    public FxpSymLink createSymLink()
    {
        return new FxpSymLink(version, nextId++);
    }

    public FxpStatus createStatus()
    {
        return new FxpStatus(version, nextId++);
    }

    public FxpHandle createHandle()
    {
        return new FxpHandle(version, nextId++);
    }

    public FxpData createData()
    {
        return new FxpData(version, nextId++);
    }

    public FxpName createName()
    {
        return new FxpName(version, nextId++);
    }

    public FxpAttrs createAttrs()
    {
        return new FxpAttrs(version, nextId++);
    }

    public FxpExtended createExtended()
    {
        return new FxpExtended(version, nextId++);
    }

    public FxpExtendedReply createExtendedReply()
    {
        return new FxpExtendedReply(version, nextId++);
    }
}