package org.klomp.sftp.protocol;

import java.io.IOException;
import java.io.InputStream;
import java.util.SortedMap;
import java.util.TreeMap;

import org.klomp.eclipse.team.sftp.SftpException;
import org.klomp.eclipse.team.sftp.client.SftpCommunicationException;
import org.klomp.sftp.packet.FxpData;
import org.klomp.sftp.packet.FxpRead;
import org.klomp.sftp.packet.FxpStatus;
import org.klomp.sftp.packet.SftpPacket;

public class SftpFileInputStream extends InputStream
{
    /**
     * The maximum number of outstanding requests for which no packets have yet been received.
     */
    public final static int MAX_REQUEST_AT_ONCE_BLOCKS = 20;

    /**
     * The maximum number of bytes of data that can be requested in a single read request to the
     * sftp server.
     */
    private final static int MAX_DATA_SIZE = 32000;

    /**
     * How many outstanding requests for which no packets have yet been received exist.
     */
    private int currentRequestAtOnceBlocks = 2;

    /**
     * The remote file from which this stream reads.
     */
    private SftpFile file;

    /**
     * Blocks that have been received but have not been (fully) read through the read methods of this
     * stream.
     */
    private SortedMap receivedBlocks = new TreeMap();

    /**
     * Keyed by packet id of request packet.
     */
    private SortedMap requestedBlocks = new TreeMap();

    /**
     * The position in the remote file of the next byte to request.
     */
    private long requestPosition;

    /**
     * The position in the stream at which the next byte will be read through this stream.
     */
    private long position;

    /**
     * Whether all packets have been received from the remote, and no more packets need to be requested.
     */
    private boolean remoteEof;

    /**
     * Whether this stream has reach the end of the remote file.
     */
    private boolean eof;

    SftpFileInputStream(SftpFile file, long position)
    {
        this.file = file;
        this.position = position;
        this.requestPosition = position;
    }

    public void close() throws IOException
    {
        if (eof)
        {
            return;
        }
        try
        {
            eof = true;
            file.close();
        }
        catch (SftpException e)
        {
            throw new SftpIOException(e);
        }
    }

    public int read() throws IOException
    {
        while (!eof)
        {
            while (!receivedBlocks.isEmpty())
            {
                Long key = (Long) receivedBlocks.firstKey();
                TransferBlock block = (TransferBlock) receivedBlocks.get(key);
                long blockPosition = block.getPosition();
                byte[] data = block.getData();
                if (position >= blockPosition)
                {
                    if (position < blockPosition + data.length)
                    {
                        byte result = data[(int) (position - blockPosition)];
                        position++;
                        return result;
                    }
                    else
                    {
                        receivedBlocks.remove(key);
                    }
                }
            }
            readFromRemote();
        }
        return -1;
    }

    public int read(byte[] b, int off, int len) throws IOException
    {
        while (!eof)
        {
            while (!receivedBlocks.isEmpty())
            {
                Long key = (Long) receivedBlocks.firstKey();
                //System.err.println("Received block key: " + key);
                TransferBlock block = (TransferBlock) receivedBlocks.get(key);
                long blockPosition = block.getPosition();
                byte[] data = block.getData();
                if (position >= blockPosition)
                {
                    if (position < blockPosition + data.length)
                    {
                        int blockOffset = (int) (position - blockPosition);
                        int blockReadLen = data.length - blockOffset;
                        int readLen = Math.min(len, blockReadLen);
                        System.arraycopy(data, blockOffset, b, off, readLen);
                        position += readLen;
                        //System.err.println("read(" + len + ") = " + readLen);
                        return readLen;
                    }
                    else
                    {
                        receivedBlocks.remove(key);
                    }
                }
            }
            readFromRemote();
        }
        return -1;
    }

    private void readFromRemote() throws IOException
    {
        if (remoteEof && receivedBlocks.isEmpty())
        {
//            System.err.println(
//                "remoteEof, requestedBlocks.size() == "
//                    + requestedBlocks.size()
//                    + ", receivedBlocks.size() == "
//                    + receivedBlocks.size());
            close();
            return;
        }
        try
        {
            fillGap();
            requestMoreBlocks();
            readResponse();
        }
        catch (SftpException e)
        {
            throw new SftpIOException(e);
        }
    }

    private void readResponse() throws SftpException
    {
        Integer packetId = (Integer) requestedBlocks.firstKey();
        SftpPacket response = file.connection.readPacket(packetId.intValue());
        switch (response.getType())
        {
            case SftpPacket.FXP_STATUS :
                FxpStatus status = (FxpStatus) response;
                if (status.getCode() == FxpStatus.FX_EOF)
                {
                    Integer lastRequestedPacketId = (Integer) requestedBlocks.lastKey();
                    file.connection.setMaxIgnorePacketId(lastRequestedPacketId.intValue());
                    remoteEof = true;
                    break;
                }
                file.defaultStatusHandler((FxpStatus) response);
                throw new SftpException("Unexpected non-error status received: " + response);
            case SftpPacket.FXP_DATA :
                TransferBlock block = (TransferBlock) requestedBlocks.get(packetId);
                if (block == null)
                {
                    throw new SftpException("Packet Id inconsistency - internal error");
                }
                FxpData data = (FxpData) response;
                block.setData(data.getData());
                receivedBlocks.put(new Long(block.getPosition()), block);
                requestedBlocks.remove(packetId);
                break;
            default :
                throw new SftpException("Unexpected packet received: " + response);
        }
    }

    private void requestMoreBlocks() throws SftpCommunicationException
    {
        int gap = currentRequestAtOnceBlocks - requestedBlocks.size();
        if (gap >= (currentRequestAtOnceBlocks / 2))
        {
            for (int i = 0; i < gap; i++)
            {
                requestBlock(requestPosition, MAX_DATA_SIZE);
                requestPosition += MAX_DATA_SIZE;
            }
            currentRequestAtOnceBlocks = Math.min(MAX_REQUEST_AT_ONCE_BLOCKS, currentRequestAtOnceBlocks + 1);
        }
    }

    private void fillGap() throws SftpCommunicationException
    {
        if (!receivedBlocks.isEmpty())
        {
            Long key = (Long) receivedBlocks.firstKey();
            TransferBlock block = (TransferBlock) receivedBlocks.get(key);
            long blockPosition = block.getPosition();
            if (position < blockPosition)
            {
                // There is a gap between the first received block in the queue and the read
                // position. Request the data in the gap, so it will become available.
                requestBlock(position, Math.min(MAX_DATA_SIZE, (int) (blockPosition - position)));
            }
        }
    }

    private int requestBlock(long offset, int length) throws SftpCommunicationException
    {
        FxpRead pkt = file.pktFactory.createRead();
        pkt.setHandle(file.getHandle());
        pkt.setLength(length);
        pkt.setOffset(offset);

        file.connection.sendPacket(pkt);

        TransferBlock block = new TransferBlock(offset);
        requestedBlocks.put(new Integer(pkt.getId()), block);
        return pkt.getId();
    }
}
