package org.klomp.ssh;

import java.io.ByteArrayOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;

/**
 * 
 */
public class DataIO
{

    public static void setInt(byte[] b, int offset, int value)
    {
        b[offset + 0] = (byte) (value >>> 24);
        b[offset + 1] = (byte) (value >>> 16);
        b[offset + 2] = (byte) (value >>> 8);
        b[offset + 3] = (byte) value;
    }

    public static void writeBoolean(ByteArrayOutputStream out, boolean value)
    {
        out.write(value ? 1 : 0);
    }

    public static void writeInt(ByteArrayOutputStream out, int value)
    {
        out.write(value >>> 24);
        out.write(value >>> 16);
        out.write(value >>> 8);
        out.write(value);
    }

    public static void writeInt(PrintStream out, int value)
    {
        out.write(value >>> 24);
        out.write(value >>> 16);
        out.write(value >>> 8);
        out.write(value);
    }

    public static void writeLong(ByteArrayOutputStream out, long value)
    {
        out.write((byte) (value >>> 56));
        out.write((byte) (value >>> 48));
        out.write((byte) (value >>> 40));
        out.write((byte) (value >>> 32));

        out.write((byte) (value >>> 24));
        out.write((byte) (value >>> 16));
        out.write((byte) (value >>> 8));
        out.write((byte) value);
    }

    /**
     * Always writes <code>size</code> bytes from the array <code>value</code>. If the array is too small,
     * extra zero-value bytes are written until a total of <code>size</code> bytes have been output.
     */
    public static void writeBytes(ByteArrayOutputStream out, byte[] value, int size)
    {
        out.write(value, 0, Math.min(value.length, size));
        for (int i = size - value.length; i > 0; i--)
        {
            out.write(0);
        }
    }

    public static void writeBigInteger(ByteArrayOutputStream out, BigInteger value)
    {
        writeString(out, value.toByteArray());
    }

    public static void writeString(ByteArrayOutputStream out, byte[] value)
    {
        writeInt(out, value.length);
        out.write(value, 0, value.length);
    }

    public static void writeString(ByteArrayOutputStream out, byte[] value, int offset, int length)
    {
        writeInt(out, length);
        out.write(value, offset, length);
    }

    public static void writeStringUTF8(ByteArrayOutputStream out, String value)
    {
        byte[] string;
        try
        {
            string = value.getBytes("UTF-8");
        }
        catch (UnsupportedEncodingException e)
        {
            throw new Error("Required UTF-8 encoding not supported");
        }
        writeInt(out, string.length);
        out.write(string, 0, string.length);
    }

    public static void writeStringAscii(ByteArrayOutputStream out, String value)
    {
        byte[] string;
        try
        {
            string = value.getBytes("US-Ascii");
        }
        catch (UnsupportedEncodingException e)
        {
            throw new Error("Required US-Ascii encoding not supported");
        }
        writeInt(out, string.length);
        out.write(string, 0, string.length);
    }

    public static int getInt(byte[] b, int offset)
    {
        return ((b[offset + 0] & 0xff) << 24)
            | ((b[offset + 1] & 0xff) << 16)
            | ((b[offset + 2] & 0xff) << 8)
            | (b[offset + 3] & 0xff);
    }

    public static String getString(byte[] b, int offset)
    {
        int length = getInt(b, offset);
        try
        {
            return new String(b, offset + 4, length, "UTF-8");
        }
        catch (UnsupportedEncodingException e)
        {
            throw new Error("Required UTF-8 encoding not supported");
        }
    }

    public static boolean readBoolean(InputStream in) throws IOException
    {
        int b1 = in.read();
        if (b1 == -1)
            throw new EOFException();
        return b1 != 0;
    }

    public static int readByte(InputStream in) throws IOException
    {
        int b = in.read();
        if (b == -1)
            throw new EOFException();
        return b;
    }

    public static int readInt(InputStream in) throws IOException
    {
        int b1 = in.read();
        if (b1 == -1)
            throw new EOFException();
        int b2 = in.read();
        if (b2 == -1)
            throw new EOFException();
        int b3 = in.read();
        if (b3 == -1)
            throw new EOFException();
        int b4 = in.read();
        if (b4 == -1)
            throw new EOFException();

        //                System.err.println("Readint:   " + b1 + ", " + b2 + ", " + b3 + ", " + b4);
        //                System.err.println(
        //                    "Readint/c: "
        //                        + (char) b1
        //                        + ", "
        //                        + (char) b2
        //                        + ", "
        //                        + (char) b3
        //                        + ", "
        //                        + (char) b4);
        return ((b1 & 0xff) << 24) | ((b2 & 0xff) << 16) | ((b3 & 0xff) << 8) | (b4 & 0xff);
    }

    public static long readLong(InputStream in) throws IOException
    {
        int b1 = in.read();
        if (b1 == -1)
            throw new EOFException();
        int b2 = in.read();
        if (b2 == -1)
            throw new EOFException();
        int b3 = in.read();
        if (b3 == -1)
            throw new EOFException();
        int b4 = in.read();
        if (b4 == -1)
            throw new EOFException();
        int b5 = in.read();
        if (b5 == -1)
            throw new EOFException();
        int b6 = in.read();
        if (b6 == -1)
            throw new EOFException();
        int b7 = in.read();
        if (b7 == -1)
            throw new EOFException();
        int b8 = in.read();
        if (b8 == -1)
            throw new EOFException();

        return ((b1 & 0xffL) << 56)
            | ((b2 & 0xffL) << 48)
            | ((b3 & 0xffL) << 40)
            | ((b4 & 0xffL) << 32)
            | ((b5 & 0xffL) << 24)
            | ((b6 & 0xffL) << 16)
            | ((b7 & 0xffL) << 8)
            | (b8 & 0xffL);
    }

    public static BigInteger readBigInteger(InputStream in) throws IOException
    {
        byte[] value = readString(in);
        if (value.length == 0)
        {
            return BigInteger.ZERO;
        }
        else
        {
            return new BigInteger(value);
        }
    }

    public static byte[] readString(InputStream in) throws IOException
    {
        int length = readInt(in);
        byte[] string = new byte[length];
        readFully(in, string);
        return string;
    }

    public static String readStringAscii(InputStream in) throws IOException
    {
        int length = readInt(in);
        byte[] string = new byte[length];
        readFully(in, string);
        try
        {
            return new String(string, "US-Ascii");
        }
        catch (UnsupportedEncodingException e)
        {
            throw new Error("Required US-Ascii encoding not supported");
        }
    }

    public static String readStringUTF8(InputStream in) throws IOException
    {
        int length = readInt(in);
        byte[] string = new byte[length];
        readFully(in, string);
        try
        {
            return new String(string, "UTF-8");
        }
        catch (UnsupportedEncodingException e)
        {
            throw new Error("Required UTF-8 encoding not supported");
        }
    }

    public static void readFully(InputStream in, byte[] bytes) throws IOException
    {
        readFully(in, bytes, 0, bytes.length);
    }

    public static void readFully(InputStream in, byte[] bytes, int offset, int size)
        throws IOException
    {
        for (int i = 0; i < size;)
        {
            int rs = in.read(bytes, i + offset, size - i);
            if (rs <= 0)
                throw new EOFException();
            i += rs;
        }
    }

    public static void skipFully(InputStream in, int bytes) throws IOException
    {
        for (int i = 0; i < bytes;)
        {
            int rs = (int) in.skip(bytes - i);
            if (rs <= 0)
                throw new EOFException();
            i += rs;
        }
    }
}
